package ssw.mj.test;

import java.io.StringReader;

import ssw.mj.Parser;
import ssw.mj.Scanner;
import ssw.mj.symtab.Obj;
import ssw.mj.symtab.Struct;
import ssw.mj.symtab.Tab;

/**
* Test cases for the <code>Parser</code> class.
*
* @author AW
*/
public class ParserTest extends IdentifiedTestCase {
	private static final String CR = "\r";
	private static final String LF = "\n";

	private static TestPrintWriter output;

	//---------- helpers ----------

	private void initParser (String program) {
		output = new TestPrintWriter();
		Scanner.init(new StringReader(program), null);
	}

	//--------------------------------
	//---------- test cases ----------

	/** Test the shortest syntactically valid program: */
	public void testShortestProgram () {
		initParser("program Shortest { }");

		Parser.parse(output);
		output.addExpectedLine("-- line 1, col 20: main", "NOT_FOUND", "");
		output.verify();

		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		// just to make sure that the Universe is also there
		actual = Tab.find("null");
		expected = new Obj(Obj.Con, "null", Tab.nullType);
		expected.val = 0;
		assertEquals("Universe 'null'", expected, actual);

		// only name in symbol table of this program
		actual = Tab.find("Shortest");
		expected = new Obj(Obj.Prog, "Shortest", Tab.noType);
		expected.locals = null;
		assertEquals("Program 'Shortest'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	public void testDoNothingProgram () {
		initParser(	"program DoNothing" + LF +
					"\tint dummy;" + LF +
					"{" + LF +
					"\tint doNothing() { dummy=0; return dummy; }" + LF +
					"" + LF +
					"\tvoid main() { " + LF +
					"\t\tdummy= doNothing();" + LF +
					"\t}" + LF +
					"}");

		Parser.parse(output);
		output.verify();

		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		actual = Tab.find("DoNothing");
		expected = new Obj(Obj.Prog, "DoNothing", Tab.noType);
		// we cheat here, because we will compare all the rest later
		expected.locals = actual.locals;
		assertEquals("Program 'DoNothing'", expected, actual);

		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately
		// NOTE: thus we get better error messages

		actual = actual.locals;
		expected = new Obj(Obj.Var, "dummy", Tab.intType);
		expected.adr = 0;
		expected.level = 0;
		assertEquals("Global 'dummy'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Meth, "doNothing", Tab.intType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'doNothing'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);

	}

	public void testRefTypeComparison () {
		initParser(
			"program RTC" + LF +
			"  int[] iarr, jarr;" + LF +
			"{" + LF +
			"  void test () {" + LF +
			"    if (iarr == null);" + LF +
			"    if (iarr != null);" + LF +
			"    if (iarr == jarr);" + LF +
			"    if (iarr != jarr);" + LF +
			"  }" + LF +
			"  void main () { }" + LF +
			"}"
		);

		Parser.parse(output);
		output.verify();
		
		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		actual = Tab.find("RTC");
		expected = new Obj(Obj.Prog, "RTC", Tab.noType);
		// we cheat here, because we will compare all the rest later
		expected.locals = actual.locals;
		assertEquals("Program 'RTC'", expected, actual);

		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately
		// NOTE: thus we get better error messages

		actual = actual.locals;
		expected = new Obj(Obj.Var, "iarr", new Struct(Struct.Arr, Tab.intType));
		expected.adr = 0;
		expected.level = 0;
		assertEquals("Global 'iarr'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Var, "jarr", new Struct(Struct.Arr, Tab.intType));
		expected.adr = 1;
		expected.level = 0;
		assertEquals("Global 'jarr'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Meth, "test", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'test'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	/* suggested by KBP */
	public void testTwoReturns () {
		initParser("program TR {" + LF +
		           "  void main () {}" + LF +
		           "  int getInt (int x) {" + LF +
		           "    if (x > 5) return 0;" + LF +
		           "    if (x > 0) return 1;" + LF +
		           "    return 2;" + LF +
		           "  }" + LF +
		           "}");

		Parser.parse(output);
		output.verify();
		
		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		actual = Tab.find("TR");
		expected = new Obj(Obj.Prog, "TR", Tab.noType);
		// we cheat here, because we will compare all the rest later
		expected.locals = actual.locals;
		assertEquals("Program 'TR'", expected, actual);

		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately
		// NOTE: thus we get better error messages

		actual = actual.locals;
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'main'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Meth, "getInt", Tab.intType);
		expected.adr = 0;
		expected.nPars = 1;
		Obj expectedParam = new Obj(Obj.Var, "x", Tab.intType);
		expectedParam.level = 1;
		expected.locals = expectedParam;
		assertEquals("Method 'getInt'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	/* suggest by KBP */
	public void testNestedLoops () {
		initParser(
				"program NL {" + LF +
				"  void main () int i; {" + LF +
				"    i = 0;" + LF +
				"    while (i < 10) {" + LF +
				"      while (i < 5) {" + LF +
				"        i++;" + LF +
				"        if (i == 3) break;" + LF +
				"      }" + LF +
				"      i++;" + LF +
				"    }" + LF +
				"  }" + LF +
				"}"
		);

		Parser.parse(output);
		output.verify();
		
		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		actual = Tab.find("NL");
		expected = new Obj(Obj.Prog, "NL", Tab.noType);
		// we cheat here, because we will compare all the rest later
		expected.locals = actual.locals;
		assertEquals("Program 'NL'", expected, actual);

		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately
		// NOTE: thus we get better error messages

		actual = actual.locals;
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		Obj expectedLocal = new Obj(Obj.Var, "i", Tab.intType);
		expectedLocal.level = 1;
		expected.locals = expectedLocal;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	/* suggest by KBP */
	public void testFuncAsProc () {
		initParser("program FAP {" + LF +
							 "  int getInt () { return 1; }" + LF +
							 "  void main () {" + LF +
							 "    getInt();" + LF +
							 "  }" + LF +
							 "}");

		Parser.parse(output);
		output.verify();
		
		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected;

		actual = Tab.find("FAP");
		expected = new Obj(Obj.Prog, "FAP", Tab.noType);
		// we cheat here, because we will compare all the rest later
		expected.locals = actual.locals;
		assertEquals("Program 'FAP'", expected, actual);

		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately
		// NOTE: thus we get better error messages

		actual = actual.locals;
		expected = new Obj(Obj.Meth, "getInt", Tab.intType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'getInt'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	public void testScriptExample () {
		initParser( "program P" + LF +
		            "" + LF +
		            "  final int size = 10;" + LF +
		            "" + LF +
		/*line  5*/ "  class Table {" + LF +
		            "    int[] pos;" + LF +
		            "    int[] neg;" + LF +
		            "  }" + LF +
		            "" + LF +
		/*line 10*/ "  Table val;" + LF +
		            "{" + LF +
		            "  void main()" + LF +
		            "    int x, i;" + LF +
		            "  { /*---------- Initialize val */" + LF +
		/*line 15*/ "    val = new Table;" + LF +
		            "    val.pos = new int[size]; val.neg = new int[size];" + LF +
		            "    i = 0;" + LF +
		            "    while (i < size) {" + LF +
		            "      val.pos[i] = 0; val.neg[i] = 0;" + LF +
		/*line 20*/ "      i++;" + LF +
		            "    }" + LF +
		            "    /*---------- Read values */" + LF +
		            "    read(x);" + LF +
		            "    while (x > 0) {" + LF +
		/*line 25*/ "      if (0 <= x && x < size) {" + LF +
		            "        val.pos[x]++;" + LF +
		            "      } else if (-size < x && x < 0) {" + LF +
		            "        val.neg[-x]++;" + LF +
		            "      }" + LF +
		/*line 30*/ "      read(x);" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "}");

		Parser.parse(output);
		output.verify();

		////////////////////////
		// check symbol table //
		////////////////////////
		Obj actual, expected, locals;
		Struct tableT;

		actual = Tab.find("P");  // class P
		expected = new Obj(Obj.Prog, "P", Tab.noType);
		// We cheat here, because we will compare all the rest later.
		expected.locals = actual.locals;
		assertEquals("Program 'P'", expected, actual);

		// Knowing the implementation of the symbol table
		// we can traverse it and check all the names separately.
		// NOTE: This gives us more precise error messages
		//       than just re-creating the whole symbol table and
		//       have it check by one call to equals(Prog).

		actual = actual.locals;  // final int size = 10;
		expected = new Obj(Obj.Con, "size", Tab.intType);
		expected.val = 10;
		assertEquals("Constant 'size'", expected, actual);

		actual = actual.next;  // class Table { int[] pos; int[] neg; }
		tableT = new Struct(Struct.Class);
		tableT.n = 2;
		// No cheating here, field list is small enough.
		// So we explicitly re-create it and check it at once (with equals).
		expected = new Obj(Obj.Var, "pos", new Struct(Struct.Arr, Tab.intType));
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Field 'pos'", expected, actual.type.fields);
		tableT.fields = expected;
		expected = new Obj(Obj.Var, "neg", new Struct(Struct.Arr, Tab.intType));
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Field 'neg'", expected, actual.type.fields.next);
		tableT.fields.next = expected;
		expected = new Obj(Obj.Type, "Table", tableT);
		assertEquals("Type 'Table'", expected, actual);

		actual = actual.next;  // Table val;
		expected = new Obj(Obj.Var, "val", tableT);
		assertEquals("Global 'val'", expected, actual);

		actual = actual.next;  // void main () int x, i;
		expected = new Obj(Obj.Var, "x", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Local 'x'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Var, "i", Tab.intType);
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Local 'i'", expected, actual.locals.next);
		locals.next = expected;
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = locals;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	public void testAllProds () {
		initParser(	"program AllProds" + LF +
		            "  final int nPoints = 2;" + LF +
		            "  final char exclaim = '!';" + LF +
		            "" + LF +
		/*line  5*/	"  class Point {" + LF +
		            "    int x, y;" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  char[] msg;" + LF +
		/*line 10*/	"  Point[] pArr;" + LF +
		            "" + LF +
		            "{" + LF +
		            "  void printError ()" + LF +
		            "    int i;" + LF +
		/*line 15*/	"  {" + LF +
		            "    /* print error message */" + LF +
		            "    i = 0;" + LF +
		            "    while(i<len(msg)){print(msg[i]);i++;}" + LF +
		            "    print(exclaim);" + LF +
		/*line 20*/	"    do {" + LF +
		            "      i--;" + LF +
		            "      if (i < 5) break;" + LF +
		            "      print(msg[i], 2);" + LF +
		            "    } while(i > 0);" + LF +
		/*line 25*/	"  }" + LF +
		            "" + LF +
		            "  int getX (Point p) { return p.x; }" + LF +
		            "" + LF +
		            "  void setX (int x, Point p) {" + LF +
		/*line 30*/	"    if (x > -10 && x <= 9 ||" + LF +
		            "        x == 4*3/4%4+12-11*(2147483647/(-2147483647-1)%2) )" + LF +
		            "      p.x = x;" + LF +
		            "    else printError();" + LF +
		            "  }" + LF +
		/*line 35*/	"" + LF +
		            "" + LF +
		            "  void setMsg () {" + LF +
		            "    msg = new char[5];" + LF +
		            "    msg[0] = 'E';" + LF +
		/*line 40*/	"    msg[1] = 'r';" + LF +
		            "    msg[2] = 'r';" + LF +
		            "    msg[3] = 'o';" + LF +
		            "    msg[4] = 'r';" + LF +
		            "  }" + LF +
		/*line 45*/	"" + LF +
		            "  void setPoints ()" + LF +
		            "    int i;" + LF +
		            "  {" + LF +
		            "    pArr = new Point[nPoints];" + LF +
		/*line 50*/	"    i = 0;" + LF +
		            "    while (i < len(pArr)) {" + LF +
		            "      pArr[i] = new Point;" + LF +
		            "      pArr[i].x = i;" + LF +
		            "      pArr[i].y = i;" + LF +
		/*line 55*/	"      i++;" + LF +
		            "    }" + LF +
		            "  }" + LF +
		            "" + LF +
		            "  void main ()" + LF +
		/*line 60*/	"    int i;" + LF +
		            "  {" + LF +
		            "    setMsg();" + LF +
		            "    setPoints();" + LF +
		            "    read(i);" + LF +
		/*line 65*/	"    if (i < nPoints) {" + LF +
		            "      print(getX(pArr[i]));" + LF +
		            "      print(pArr[i].y);" + LF +
		            "    }" + LF +
		            "    else" + LF +
		/*line 70*/	"      printError();" + LF +
		            "    print('\\n');" + LF +
		            "  }" + LF +
		            "}");

		Parser.parse(output);
		output.verify();

		////////////////////////
		// check symbol table //
		////////////////////////

		Obj actual, expected, locals;
		Struct pointT;

		actual = Tab.find("AllProds");
		expected = new Obj(Obj.Prog, "AllProds", Tab.noType);
		// We cheat here, because we will compare all the rest later.
		expected.locals = actual.locals;
		assertEquals("Program 'AllProds'", expected, actual);

		actual = actual.locals;  // final int nPoints = 2;
		expected = new Obj(Obj.Con, "nPoints", Tab.intType);
		expected.val = 2;
		assertEquals("Constant 'nPoints'", expected, actual);

		actual = actual.next;  // final char exclaim = '!';
		expected = new Obj(Obj.Con, "exclaim", Tab.charType);
		expected.val = 33;
		assertEquals("Constant 'exclaim'", expected, actual);

		actual = actual.next;  // class Point { int x; int y; }
		pointT = new Struct(Struct.Class);
		pointT.n = 2;
		expected = new Obj(Obj.Var, "x", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Field 'x'", expected, actual.type.fields);
		pointT.fields = expected;
		expected = new Obj(Obj.Var, "y", Tab.intType);
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Field 'y'", expected, actual.type.fields.next);
		pointT.fields.next = expected;
		expected = new Obj(Obj.Type, "Point", pointT);
		assertEquals("Class 'Point'", expected, actual);

		actual = actual.next;  // char[] msg;
		expected = new Obj(Obj.Var, "msg", new Struct(Struct.Arr, Tab.charType));
		expected.adr = 0;
		expected.level = 0;
		assertEquals("Global 'msg'", expected, actual);

		actual = actual.next;
		expected = new Obj(Obj.Var, "pArr", new Struct(Struct.Arr, pointT));
		expected.adr = 1;
		expected.level = 0;
		assertEquals("Global 'pArr'", expected, actual);

		actual = actual.next;  // void printError () int i;
		expected = new Obj(Obj.Var, "i", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Local 'i'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Meth, "printError", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = locals;
		assertEquals("Method 'printError'", expected, actual);

		actual = actual.next;  // int getX (Point p)
		expected = new Obj(Obj.Var, "p", pointT);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Param 'p'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Meth, "getX", Tab.intType);
		expected.adr = 0;
		expected.nPars = 1;
		expected.locals = locals;
		assertEquals("Method 'getX'", expected, actual);

		actual = actual.next;  // void setX (int x, Point p)
		expected = new Obj(Obj.Var, "x", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Param 'x'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Var, "p", pointT);
		expected.adr = 1;
		expected.level = 1;
		assertEquals("Param 'p'", expected, actual.locals.next);
		locals.next = expected;
		expected = new Obj(Obj.Meth, "setX", Tab.noType);
		expected.adr = 0;
		expected.nPars = 2;
		expected.locals = locals;
		assertEquals("Method 'setX'", expected, actual);

		actual = actual.next;  // void setMsg ()
		expected = new Obj(Obj.Meth, "setMsg", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'setMsg'", expected, actual);

		actual = actual.next;  // void setPoints () int i;
		expected = new Obj(Obj.Var, "i", Tab.intType);
		expected.adr = 0;
		expected.level = 1;
		assertEquals("Local 'i'", expected, actual.locals);
		locals = expected;
		expected = new Obj(Obj.Meth, "setPoints", Tab.noType);
		expected.adr = 0;
		expected.level = 0;
		expected.locals = locals;
		assertEquals("Method 'setPoints'", expected, actual);

		actual = actual.next;  // void main () int i;
		// locals are the same as above
		assertEquals("Local 'i'", locals, actual.locals);
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = locals;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	// -----------------------------------------------
	// ---------- test programs with errors ----------

	public void testWrongGlobalDecl () {
		initParser("program WGD" + LF +
		           "  final int MAX = 10;" + LF +
		           "  123, i, x;" + LF +
		           "  char c1, c2, c3;" + LF +
		           "{ void main () { } }");

		Parser.parse(output);
		output.addExpectedLine("-- line 3, col 3: ", "INVALID_DECL");
		output.verify();
		assertEquals("error count", 1, Parser.Errors.count());

		////////////////////////
		// check symbol table //
		////////////////////////

		Obj actual, expected;

		actual = Tab.find("WGD");
		expected = new Obj(Obj.Prog, "WGD", Tab.noType);
		// We cheat here, because we will compare all the rest later.
		expected.locals = actual.locals;
		assertEquals("Program 'WGD'", expected, actual);

		actual = actual.locals;  // final int MAX = 10;
		expected = new Obj(Obj.Con, "MAX", Tab.intType);
		expected.val = 10;
		assertEquals("Constant 'MAX'", expected, actual);

		actual = actual.next;  // char c1;
		expected = new Obj(Obj.Var, "c1", Tab.charType);
		expected.adr = 0;
		expected.level = 0;
		assertEquals("Global 'c1'", expected, actual);

		actual = actual.next;  // char c2;
		expected = new Obj(Obj.Var, "c2", Tab.charType);
		expected.adr = 1;
		expected.level = 0;
		assertEquals("Global 'c2'", expected, actual);

		actual = actual.next;  // char c3;
		expected = new Obj(Obj.Var, "c3", Tab.charType);
		expected.adr = 2;
		expected.level = 0;
		assertEquals("Global 'c3'", expected, actual);

		actual = actual.next;  // void main ()
		expected = new Obj(Obj.Meth, "main", Tab.noType);
		expected.adr = 0;
		expected.nPars = 0;
		expected.locals = null;
		assertEquals("Method 'main'", expected, actual);

		// make sure that this is the end
		assertNull(actual.next);
	}

	// -----------------------------------------------

	public static void main (String[] args) {
		junit.textui.TestRunner.run(ParserTest.class);
	}
}